##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""File-based ZODB storage

Files are arranged as follows.

  - The first 1024 bytes are a storage metadata section.

  In this section, the first two bytes are the characters F and S.

  The next two bytes are a storage format version id, currently "42".

  The next section is a four-byte database version string, encoded as
  byte 0: major version number
  byte 1: minor version number
  byte 2: micro version number
  byte 3: release level + serialno
  (see zodb.storage.base for details)

  The rest of the section is reserved.

A transaction record consists of:

  - 8-byte transaction id, which is also a time stamp.

  - 8-byte transaction record length - 8.

  - 1-byte status code

  - 2-byte length of user name

  - 2-byte length of description

  - 2-byte length of extension attributes

  -   user name

  -   description

  -   extension attributes

  * A sequence of data records

  - 8-byte redundant transaction length -8

A data record consists of

  - 8-byte oid.

  - 8-byte serial, which is a type stamp that matches the
    transaction timestamp.

  - 8-byte previous-record file-position.

  - 8-byte beginning of transaction record file position.

  - 2-byte version length

  - 4-byte number of object references (oids)

  - 8-byte data length

  ? 8-byte position of non-version data
    (if version length > 0)

  ? 8-byte position of previous record in this version
    (if version length > 0)

  ? version string (if version length > 0)

  ? reference oids (length == # of oids * 8)

  ? data (if data length > 0)

  ? 8-byte position of data record containing data
    (data length == 0)

Note that the lengths and positions are all big-endian.
Also, the object ids time stamps are big-endian, so comparisons
are meaningful.

Version handling

  There isn't a separate store for versions.  Each record has a
  version field, indicating what version it is in.  The records in a
  version form a linked list.  Each record that has a non-empty
  version string has a pointer to the previous record in the version.
  Version back pointers are retained *even* when versions are
  committed or aborted or when transactions are undone.

  There is a notion of "current" version records, which are the
  records in a version that are the current records for their
  respective objects.  When a version is comitted, the current records
  are committed to the destination version.  When a version is
  aborted, the current records are aborted.

  When committing or aborting, we search backward through the linked
  list until we find a record for an object that does not have a
  current record in the version.  If we find a record for which the
  non-version pointer is the same as the previous pointer, then we
  forget that the corresponding object had a current record in the
  version. This strategy allows us to avoid searching backward through
  previously committed or aborted version records.

  Of course, we ignore records in undone transactions when committing
  or aborting.

Backpointers

  When we commit or abort a version, we don't copy (or delete)
  and data.  Instead, we write records with back pointers.

  A version record *never* has a back pointer to a non-version
  record, because we never abort to a version.  A non-version record
  may have a back pointer to a version record or to a non-version
  record.

$Id: file.py,v 1.21 2003/03/17 20:23:47 bwarsaw Exp $
"""

from __future__ import generators

import os
import sys
import time
import errno
import base64
import struct
import logging
from struct import pack, unpack
from cPickle import Pickler, Unpickler, loads

try:
    from posix import fsync
except:
    fsync = None

import zodb.db
from zodb.storage.base import BaseStorage, splitrefs
from zodb import conflict
from zodb import interfaces
from zodb.interfaces import _fmt_oid
from zodb.interfaces import *
from zodb.timestamp import TimeStamp, newTimeStamp, timeStampFromTime
from zodb.lockfile import LockFile
from zodb.utils import p64, u64, cp
from zodb.storage.fsindex import fsIndex
from zodb.storage.interfaces import *

t32 = 1L << 32
# the struct formats for the headers
TRANS_HDR = ">8sQcHHH"
DATA_HDR = ">8s8sQQHIQ"
# constants to support various header sizes
TRANS_HDR_LEN = 23
DATA_HDR_LEN = 46
DATA_VERSION_HDR_LEN = 62
assert struct.calcsize(TRANS_HDR) == TRANS_HDR_LEN
assert struct.calcsize(DATA_HDR) == DATA_HDR_LEN

logger = logging.getLogger("zodb.storage.file")

warn = logger.warn
error = logger.error
nearPanic = logger.critical

def panic(message, *data):
    logger.critical(message, *data)
    raise CorruptedTransactionError(message % data)

class FileStorageError(StorageError):
    pass

class PackError(FileStorageError):
    pass

class FileStorageFormatError(FileStorageError):
    """Invalid file format

    The format of the given file is not valid.
    """

class CorruptedFileStorageError(FileStorageError, StorageSystemError):
    """Corrupted file storage."""

class CorruptedTransactionError(CorruptedFileStorageError):
    pass

class CorruptedDataError(CorruptedFileStorageError):

    def __init__(self, oid=None, buf=None, pos=None):
        self.oid = oid
        self.buf = buf
        self.pos = pos

    def __str__(self):
        if self.oid:
            msg = "Error reading oid %s.  Found %r" % (_fmt_oid(self.oid),
                                                       self.buf)
        else:
            msg = "Error reading unknown oid.  Found %r" % self.buf
        if self.pos:
            msg += " at %d" % self.pos
        return msg

class FileStorageQuotaError(FileStorageError, StorageSystemError):
    """File storage quota exceeded."""

class FileStorageFormatter:
    """Mixin class that can read and write the low-level format."""

    # subclasses must provide _file

    def _read_index(self, index, vindex, tindex, stop=MAXTID,
                    ltid=ZERO, start=None, maxoid=ZERO, recover=0,
                    read_only=0):
        """Scan the entire file storage and recreate the index.

        Returns file position, max oid, and last transaction id.  It also
        stores index information in the three dictionary arguments.

        Arguments:
        index -- dictionary, oid -> data record
        vindex -- dictionary, oid -> data record for version data
        tindex -- dictionary, oid -> data record
           XXX tindex is cleared before return, so it will be empty

        There are several default arguments that affect the scan or the
        return values.  XXX should document them.

        The file position returned is the position just after the last
        valid transaction record.  The oid returned is the maximum object
        id in the data.  The transaction id is the tid of the last
        transaction.
        """
        self._file.seek(0, 2)
        file_size = self._file.tell()
        self._file.seek(0)

        if start is None:
            start = self._metadata_size

        if file_size:
            if file_size < start:
                raise FileStorageFormatError(self._file.name)
            self._read_metadata()
        else:
            if not read_only:
                self._write_metadata()
            return self._metadata_size, maxoid, ltid

        pos = start
        self._file.seek(start)
        tid = '\0' * 7 + '\1'

        while True:
            # Read the transaction record
            h = self._file.read(TRANS_HDR_LEN)
            if not h:
                break
            if len(h) != TRANS_HDR_LEN:
                if not read_only:
                    warn('%s truncated at %s', self._file.name, pos)
                    self._file.seek(pos)
                    self._file.truncate()
                break

            tid, tl, status, ul, dl, el = unpack(TRANS_HDR, h)
            if el < 0:
                el = t32 - el

            if tid <= ltid:
                warn("%s time-stamp reduction at %s", self._file.name, pos)
            ltid = tid

            if pos+(tl+8) > file_size or status=='c':
                # Hm, the data were truncated or the checkpoint flag
                # wasn't cleared.  They may also be corrupted, in
                # which case, we don't want to totally lose the data.
                if not read_only:
                    warn("%s truncated, possibly due to damaged records at %s",
                         self._file.name, pos)
                    _truncate(self._file, self._file.name, pos)
                break

            if status not in ' up':
                warn('%s has invalid status, %s, at %s',
                     self._file.name, status, pos)

            if tl < (TRANS_HDR_LEN+ul+dl+el):
                # We're in trouble. Find out if this is bad data in
                # the middle of the file, or just a turd that Win 9x
                # dropped at the end when the system crashed.  Skip to
                # the end and read what should be the transaction
                # length of the last transaction.
                self._file.seek(-8, 2)
                rtl = u64(self._file.read(8))
                # Now check to see if the redundant transaction length is
                # reasonable:
                if file_size - rtl < pos or rtl < TRANS_HDR_LEN:
                    nearPanic('%s has invalid transaction header at %s',
                              self._file.name, pos)
                    if not read_only:
                        warn("It appears that there is invalid data at the "
                             "end of the file, possibly due to a system "
                             "crash.  %s truncated to recover from bad data "
                             "at end.", self._file.name)
                        _truncate(file, self._file.name, pos)
                    break
                else:
                    if recover:
                        return pos, None, None
                    panic('%s has invalid transaction header at %s',
                          self._file.name, pos)

            if tid >= stop:
                break

            tpos = pos
            tend = tpos + tl

            if status == 'u':
                # Undone transaction, skip it
                self._file.seek(tend)
                h = self._file.read(8)
                if h != stl:
                    if recover: return tpos, None, None
                    panic('%s has inconsistent transaction length at %s',
                          self._file.name, pos)
                pos = tend + 8
                continue

            pos = tpos + (TRANS_HDR_LEN + ul + dl + el)
            while pos < tend:
                # Read the data records for this transaction
                h = self._read_data_header(pos)
                dlen = h.recordlen()
                tindex[h.oid] = pos

                if h.version:
                    vindex[h.version] = pos

                if pos + dlen > tend or h.tloc != tpos:
                    if recover:
                        return tpos, None, None
                    panic("%s data record exceeds transaction record at %s",
                          self._file.name, pos)

                if index.get(h.oid, 0) != h.prev:
                    if h.prev:
                        if recover:
                            return tpos, None, None
                    error("%s incorrect previous pointer at %s: "
                          "index says %r record says %r",
                          self._file.name, pos, index.get(h.oid), h.prev)

                pos += dlen

            if pos != tend:
                if recover:
                    return tpos, None, None
                panic("%s data records don't add up at %s",
                      self._file.name, tpos)

            # Read the (intentionally redundant) transaction length
            self._file.seek(pos)
            l = u64(self._file.read(8))
            if l != tl:
                if recover:
                    return tpos, None, None
                panic("%s redundant transaction length check failed at %s",
                      self._file.name, pos)
            pos += 8

            if tindex: # avoid the pathological empty transaction case
                _maxoid = max(tindex.keys()) # in 2.2, just max(tindex)
                maxoid = max(_maxoid, maxoid)
                index.update(tindex)
                tindex.clear()

        return pos, maxoid, ltid

    _metadata_size = 1024
    _format_version = "43"

    def _read_metadata(self):
        # Read the 1K metadata block at the beginning of the storage.
        self._file.seek(0)
        fs = self._file.read(2)
        if fs != "FS":
            raise FileStorageFormatError(self._file.name)
        fsver = self._file.read(2)
        if fsver != self._format_version:
            raise FileStorageFormatError(self._file.name)
        ver = self._file.read(4)
        if ver != "\0" * 4:
            self._version = ver

    def _write_metadata(self):
        # Write the 1K metadata block at the beginning of the storage.
        self._file.seek(0)
        self._file.write("FS")
        self._file.write(self._format_version)
        # If self._version is not yet set, write all zeros.
        if self._version is not None:
            self._file.write(self._version)
        else:
            self._file.write("\0" * 4)
        # Fill the rest with null bytes
        self._file.write("\0" * (self._metadata_size - 8))

    def _read_data_header(self, pos, oid=None):
        """Return a DataHeader object for data record at pos.

        If ois is not None, raise CorruptedDataError if oid passed
        does not match oid in file.

        If there is version data, reads the version part of the header.
        If there is no pickle data, reads the back pointer.
        """
        self._file.seek(pos)
        s = self._file.read(DATA_HDR_LEN)
        if len(s) != DATA_HDR_LEN:
            raise CorruptedDataError(oid, s, pos)
        h = DataHeader.fromString(s)
        if oid is not None and oid != h.oid:
            raise CorruptedDataError(oid, s, pos)
        if h.vlen:
            s = self._file.read(16 + h.vlen)
            h.parseVersion(s)
        if not h.plen:
            h.back = u64(self._file.read(8))
        return h

    def _write_data_header(self, file, oid, serial, prev, tloc, vlen, nrefs,
                           plen):
        s = struct.pack(DATA_HDR, oid, serial, prev, tloc, vlen, nrefs, plen)
        file.write(s)

    def _write_version_header(self, file, pnv, vprev, version):
        s = struct.pack(">QQ", pnv, vprev)
        file.write(s + version)

    def _read_txn_header(self, pos):
        pass

    def _loadBack_impl(self, oid, back):
        # shared implementation used by various _loadBack methods
        while 1:
            if not back:
                # If backpointer is 0, object does not currently exist.
                raise POSKeyError(oid)
            h = self._read_data_header(back)
            refs = self._file.read(h.nrefs * 8)
            if h.plen:
                return self._file.read(h.plen), refs, h.serial, back, h.tloc
            back = h.back

    def _loadBack(self, oid, back):
        data, refs, serial, old, tloc = self._loadBack_impl(oid, back)
        return data, serial

    def _loadBackPOS(self, oid, back):
        """Return position of data record for backpointer."""
        data, refs, serial, old, tloc = self._loadBack_impl(oid, back)
        return old

    def _loadBackTxn(self, oid, back):
        """Return data, serial, and txn id for backpointer."""
        data, refs, serial, old, tloc = self._loadBack_impl(oid, back)
        self._file.seek(tloc)
        h = self._file.read(TRANS_HDR_LEN)
        tid = h[:8]
        refs = splitrefs(refs)
        return data, refs, serial, tid

    def getTxnFromData(self, oid, back):
        """Return transaction id for data at back."""
        h = self._read_data_header(back, oid)
        self._file.seek(h.tloc)
        # seek to transaction header, where tid is first 8 bytes
        return self._file.read(8)

class TxnTempFile(FileStorageFormatter):
    """Helper class used in conjunction with _tfile of FileStorage."""

    def __init__(self, afile):
        self._file = afile

class FileStorage(BaseStorage, FileStorageFormatter):
    # default pack time is 0
    _packt = ZERO

    __implements__ = IStorage, IUndoStorage, IVersionStorage

    def __init__(self, file_name, create=0, read_only=0, stop=None,
                 quota=None):

        # This check was removed from zodb3, because old versions of
        # python could lie if they didn't have large file support.
        if not os.path.exists(file_name):
            create = 1

        if read_only:
            self._is_read_only = 1
            if create:
                raise ValueError, "can't create a read-only file"
        elif stop is not None:
            raise ValueError("time-travel is only supported "
                             "in read-only mode")

        if stop is None:
            stop=MAXTID

        self._file_name = file_name
        super(FileStorage, self).__init__(file_name)
        self._initIndex()
        if not read_only:
            self._lock()
            self._tfile = open(file_name + '.tmp', 'w+b')
            self._tfmt = TxnTempFile(self._tfile)
        else:
            self._tfile = None

        self._file = None
        if not create:
            try:
                self._file = open(file_name, read_only and 'rb' or 'r+b')
            except IOError, exc:
                if exc.errno == errno.EFBIG:
                    # The file is too big to open.  Fail visibly.
                    raise
                if exc.errno == errno.ENOENT:
                    # The file doesn't exist.  Create it.
                    create = 1
                # If something else went wrong, it's hard to guess
                # what the problem was.  If the file does not exist,
                # create it.  Otherwise, fail.
                if os.path.exists(file_name):
                    raise
                else:
                    create = 1

        if self._file is None and create:
            if os.path.exists(file_name):
                os.remove(file_name)
                self._clear_index()
            self._file = open(file_name, 'w+b')
            self._write_metadata()

        self._conflict = conflict.ConflictResolver(self)
        r = self._restore_index()
        if r is not None:
            index, vindex, start, maxoid, ltid = r
            self._initIndex(index, vindex)
            self._pos, self._oid, tid = self._read_index(
                self._index, self._vindex, self._tindex, stop, ltid=ltid,
                start=start, maxoid=maxoid, read_only=read_only)
        else:
            self._pos, self._oid, tid = self._read_index(
                self._index, self._vindex, self._tindex, stop,
                read_only=read_only)
        self._ltid = tid

        # self._pos should always point just past the last
        # transaction.  During 2PC, data is written after _pos.
        # invariant is restored at tpc_abort() or tpc_finish().

        if tid is not None:
            self._ts = tid = TimeStamp(tid)
            t = newTimeStamp()
            if tid > t:
                warn("%s Database records in the future", file_name);
                if tid.timeTime() - t.timeTime() > 86400*30:
                    # a month in the future? This is bogus, use current time
                    self._ts = t

        self._quota = quota

    def _lock(self):
        self._lock_file = LockFile(self._name + '.lock')

    def _initIndex(self, index=None, vindex=None, tindex=None, tvindex=None):
        self._index = index or fsIndex()
        self._vindex = vindex or {}
        self._tindex = tindex or {}
        self._tvindex = tvindex or {}

    def _save_index(self):
        """Write the database index to a file to support quick startup."""

        index_name = self._name + '.index'
        tmp_name = index_name + '.index_tmp'

        f = open(tmp_name,'wb')
        p = Pickler(f,1)
        info = {'index': self._index, 'pos': self._pos,
                'oid': self._oid, 'vindex': self._vindex}

        p.dump(info)
        f.flush()
        f.close()
        try:
            try:
                os.remove(index_name)
            except OSError:
                pass
            os.rename(tmp_name, index_name)
        except:
            pass

    def _clear_index(self):
        index_name = self._name + '.index'
        if os.path.exists(index_name):
            try:
                os.remove(index_name)
            except OSError:
                pass

    def _restore_index(self):
        """Load database index to support quick startup."""
        try:
            f = open("%s.index" % self._name, 'rb')
        except:
            return None

        p = Unpickler(f)
        try:
            info = p.load()
        except:
            exc, err = sys.exc_info()[:2]
            warn("Failed to load database index: %s: %s", exc, err)
            return None
        index = info.get('index')
        vindex = info.get('vindex')
        pos = info.get('pos')
        oid = info.get('oid')
        if index is None or pos is None or oid is None or vindex is None:
            return None

        if pos > self._metadata_size: # otherwise storage is empty
            # Get the last transaction
            self._file.seek(pos - 8)
            tl = u64(self._file.read(8))
            pos -= tl + 8
            self._file.seek(pos)
            tid = self._file.read(8)
        else:
            tid = None

        return index, vindex, pos, oid, tid

    def close(self):
        self._file.close()
        if hasattr(self, '_lock_file'):
            self._lock_file.close()
        if self._tfile:
            self._tfile.close()
        try:
            self._save_index()
        except:
            logger.warn("Error saving storage index",
                        exc_info=sys.exc_info())
            pass # We don't care if this fails.

    def setVersion(self, version):
        self._version = version
        if not self._is_read_only:
            self._write_metadata()

    def abortVersion(self, src, transaction):
        return self.commitVersion(src, '', transaction, abort=True)

    def commitVersion(self, src, dest, transaction, abort=False):
        # We are going to commit by simply storing back pointers.
        if self._is_read_only:
            raise ReadOnlyError()
        if not (src and isinstance(src, str) and isinstance(dest, str)):
            raise interfaces.VersionCommitError('Invalid source version')

        if src == dest:
            raise interfaces.VersionCommitError(
                "Can't commit to same version: %s" % repr(src))

        if dest and abort:
            raise interfaces.VersionCommitError(
                "Internal error, can't abort to a version")

        if transaction is not self._transaction:
            raise StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            return self._commitVersion(src, dest, transaction, abort)
        finally:
            self._lock_release()

    def _commitVersion(self, src, dest, transaction, abort=False):
        # call after checking arguments and acquiring lock
        srcpos = self._vindex.get(src, 0)
        spos = p64(srcpos)
        # middle holds bytes 16:38 of a data record:
        #    pos of transaction, len of version name, data length, nrefs
        #    commit version never writes data, so data length and nrefs are
        #    always 0
        middle = struct.pack(">QHIQ", self._pos, len(dest), 0, 0)

        # recsize is the total size of a data record written by
        # commit version.  The var here stores the location of the
        # current data record.  here is incremented by recsize
        # each time we write an object record.

        # Add 8 to header size to account for backpointer.
        if dest:
            sd = p64(self._vindex.get(dest, 0))
            recsize = DATA_VERSION_HDR_LEN + 8 + len(dest)
        else:
            sd = ''
            recsize = DATA_HDR_LEN + 8

        here = self._pos + self._tfile.tell() + self._thl
        oids = []
        current_oids = {}
        if not abort:
            newserial = self._serial

        while srcpos:
            h = self._read_data_header(srcpos)
            if abort:
                # If we are aborting, the serialno in the new data
                # record should be the same as the serialno in the last
                # non-version data record.
                # XXX This might be the only time that the serialno
                # of a data record does not match the transaction id.
                self._file.seek(h.pnv)
                h_pnv = self._read_data_header(srcpos)
                newserial = h_pnv.serial

            if self._index.get(h.oid) == srcpos:
                # This is a current record!
                self._tindex[h.oid] = here
                oids.append(h.oid)
                self._tfile.write(h.oid + newserial + spos + middle)
                if dest:
                    self._tvindex[dest] = here
                    self._tfile.write(p64(h.pnv) + sd + dest)
                    sd = p64(here)

                self._tfile.write(abort and p64(h.pnv) or spos)
                # data backpointer to src data
                here += recsize

                current_oids[h.oid] = 1
            else:
                # XXX I don't understand this branch.  --jeremy
                # Hm.  This is a non-current record.  Is there a
                # current record for this oid?
                if not current_oids.has_key(h.oid):
                    break

            srcpos = h.vprev
            spos = p64(srcpos)
        return oids

    def load(self, oid, version):
        self._lock_acquire()
        try:
            try:
                pos = self._index[oid]
            except KeyError:
                raise POSKeyError(oid)
            h = self._read_data_header(pos, oid)
            if h.version and h.version != version:
                return self._loadBack(oid, h.pnv)

            # If we get here, then either this was not a version record,
            # or we've already read past the version data!  Read past any
            # references data first.
            self._file.read(h.nrefs * 8)

            if h.plen:
                return self._file.read(h.plen), h.serial
            # We use the current serial, since that is the one that
            # will get checked when we store.
            return self._loadBack(oid, h.back)[0], h.serial
        finally:
            self._lock_release()

    def loadSerial(self, oid, serial):
        self._lock_acquire()
        try:
            try:
                pos = self._index[oid]
            except KeyError:
                raise POSKeyError(oid)
            while 1:
                h = self._read_data_header(pos, oid)
                if h.serial == serial:
                    break
                # Keep looking for serial
                pos = h.prev
                if not pos:
                    # XXX serial?
                    raise POSKeyError(serial)
                continue

            # Read past any references data
            self._file.read(h.nrefs * 8)
            if h.plen:
                return self._file.read(h.plen)
            return self._loadBack(oid, h.pnv)[0]
        finally:
            self._lock_release()

    def modifiedInVersion(self, oid):
        self._lock_acquire()
        try:
            try:
                pos = self._index[oid]
            except KeyError:
                raise POSKeyError(oid)
            h = self._read_data_header(pos, oid)
            return h.version
        finally:
            self._lock_release()

    def store(self, oid, serial, data, refs, version, transaction):
        if self._is_read_only:
            raise ReadOnlyError()
        if transaction is not self._transaction:
            raise StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            old = self._index.get(oid, 0)
            pnv = None
            if old:
                h = self._read_data_header(old)
                if h.version:
                    if version != h.version:
                        raise interfaces.VersionLockError(oid, h.version)
                    pnv = h.pnv

                if serial != h.serial:
                    data, refs = self._conflict.resolve(
                        oid, h.serial, serial, data)

            pos = self._pos
            here = pos + self._tfile.tell() + self._thl
            self._tindex[oid] = here
            self._write_data_header(self._tfile,
                                    oid, self._serial, old, pos,
                                    len(version), len(refs), len(data))
            if version:
                # Link to last record for this version:
                pv = (self._tvindex.get(version, 0)
                      or self._vindex.get(version, 0))
                if pnv is None:
                    pnv = old
                self._write_version_header(self._tfile, pnv, pv, version)
                self._tvindex[version] = here

            self._tfile.write(''.join(refs))
            self._tfile.write(data)

            if self._quota is not None and here > quota:
                raise FileStorageQuotaError("storage quota exceeded")

            if old and serial != h.serial:
                return conflict.ResolvedSerial
            else:
                return self._serial

        finally:
            self._lock_release()

    def _data_find(self, tpos, oid, data):
        # Return backpointer to oid in data record for in transaction at tpos.
        # It should contain a pickle identical to data. Returns 0 on failure.
        # Must call with lock held.
        self._file.seek(tpos)
        h = self._file.read(TRANS_HDR_LEN)
        tid, tl, status, ul, dl, el = struct.unpack(TRANS_HDR, h)
        self._file.read(ul + dl + el)
        tend = tpos + tl
        pos = self._file.tell()
        while pos < tend:
            h = self._read_data_header(pos)
            if h.oid == oid:
                # Read past any references data
                self._file.read(h.nrefs * 8)
                # Make sure this looks like the right data record
                if h.plen == 0:
                    # This is also a backpointer.  Gotta trust it.
                    return pos
                if h.plen != len(data):
                    # The expected data doesn't match what's in the
                    # backpointer.  Something is wrong.
                    error("Mismatch between data and backpointer at %d", pos)
                    return 0
                _data = self._file.read(h.plen)
                if data != _data:
                    return 0
                return pos
            pos += h.recordlen()
        return 0

    def restore(self, oid, serial, data, refs, version, prev_txn, transaction):
        # A lot like store() but without all the consistency checks.  This
        # should only be used when we /know/ the data is good, hence the
        # method name.  While the signature looks like store() there are some
        # differences:
        #
        # - serial is the serial number of /this/ revision, not of the
        #   previous revision.  It is used instead of self._serial, which is
        #   ignored.
        #
        # - Nothing is returned
        #
        # - data can be None, which indicates a George Bailey object
        #   (i.e. one who's creation has been transactionally undone).
        #
        # prev_txn is a backpointer.  In the original database, it's possible
        # that the data was actually living in a previous transaction.  This
        # can happen for transactional undo and other operations, and is used
        # as a space saving optimization.  Under some circumstances the
        # prev_txn may not actually exist in the target database (i.e. self)
        # for example, if it's been packed away.  In that case, the prev_txn
        # should be considered just a hint, and is ignored if the transaction
        # doesn't exist.
        if self._is_read_only:
            raise ReadOnlyError()
        if transaction is not self._transaction:
            raise StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            prev_pos = 0
            if prev_txn is not None:
                prev_txn_pos = self._txn_find(prev_txn, 0)
                if prev_txn_pos:
                    prev_pos = self._data_find(prev_txn_pos, oid, data)
            old = self._index.get(oid, 0)
            # Calculate the file position in the temporary file
            here = self._pos + self._tfile.tell() + self._thl
            # And update the temp file index
            self._tindex[oid] = here
            if prev_pos:
                # If there is a valid prev_pos, don't write data.
                data = None
            if data is None:
                dlen = 0
                refs = []
            else:
                dlen = len(data)
            # Write the recovery data record
            self._tfile.write(pack(DATA_HDR,
                                   oid, serial, old, self._pos,
                                   len(version), len(refs), dlen))
            # We need to write some version information if this revision is
            # happening in a version.
            if version:
                pnv = self._restore_pnv(oid, old, version, prev_pos)
                if pnv is not None:
                    self._tfile.write(p64(pnv))
                else:
                    self._tfile.write(p64(old))
                # Link to the last record for this version
                pv = self._tvindex.get(version, 0)
                if not pv:
                    pv = self._vindex.get(version, 0)
                self._tfile.write(p64(pv))
                self._tvindex[version] = here
                self._tfile.write(version)
            # Write the references
            self._tfile.write(''.join(refs))
            # Write the data or a backpointer
            if data is None:
                if prev_pos:
                    self._tfile.write(p64(prev_pos))
                else:
                    # Write a zero backpointer, which indicates an
                    # un-creation transaction.
                    self._tfile.write(ZERO)
            else:
                self._tfile.write(data)
        finally:
            self._lock_release()

    def _restore_pnv(self, oid, prev, version, bp):
        # Find a valid pnv (previous non-version) pointer for this version.

        # If there is no previous record, there can't be a pnv.
        if not prev:
            return None

        pnv = None
        h = self._read_data_header(prev, oid)
        # If the previous record is for a version, it must have
        # a valid pnv.
        if h.version:
            return h.pnv
        elif bp:
            # XXX Not sure the following is always true:
            # The previous record is not for this version, yet we
            # have a backpointer to it.  The current record must
            # be an undo of an abort or commit, so the backpointer
            # must be to a version record with a pnv.
            h2 = self._read_data_header(bp, oid)
            if h2.version:
                return h2.pnv
            else:
                warn("restore could not find previous non-version data "
                     "at %d or %d", prev, bp)
                return None

    def _clear_temp(self):
        self._tindex.clear()
        self._tvindex.clear()
        if self._tfile is not None:
            self._tfile.seek(0)

    def _begin(self, tid, u, d, e):
        self._nextpos = 0
        self._thl = TRANS_HDR_LEN + len(u) + len(d) + len(e)
        if self._thl > 65535:
            # one of u, d, or e may be > 65535
            # We have to check lengths here because struct.pack
            # doesn't raise an exception on overflow!
            if len(u) > 65535:
                raise FileStorageError('user name too long')
            if len(d) > 65535:
                raise FileStorageError('description too long')
            if len(e) > 65535:
                raise FileStorageError('too much extension data')


    def tpcVote(self, transaction):
        self._lock_acquire()
        try:
            if transaction is not self._transaction:
                return
            dlen = self._tfile.tell()
            if not dlen:
                return # No data in this trans
            self._tfile.seek(0)
            user, desc, ext = self._ude

            self._file.seek(self._pos)
            tl = self._thl + dlen

            try:
                # Note that we use a status of 'c', for checkpoint.
                # If this flag isn't cleared, anything after this is
                # suspect.
                self._file.write(pack(">8sQcHHH",
                                      self._serial, tl, "c",
                                      len(user), len(desc), len(ext)))
                if user:
                    self._file.write(user)
                if desc:
                    self._file.write(desc)
                if ext:
                    self._file.write(ext)

                cp(self._tfile, self._file, dlen)

                self._file.write(p64(tl))
                self._file.flush()
            except:
                # Hm, an error occured writing out the data. Maybe the
                # disk is full. We don't want any turd at the end.
                self._file.truncate(self._pos)
                raise
            self._nextpos = self._pos + (tl + 8)
        finally:
            self._lock_release()

    def _finish(self, tid, u, d, e):
        nextpos = self._nextpos
        if nextpos:
            # Clear the checkpoint flag
            self._file.seek(self._pos +16 )
            self._file.write(self._tstatus)
            self._file.flush()

            if fsync is not None:
                fsync(self._file.fileno())

            self._pos = nextpos

            self._index.update(self._tindex)
            self._vindex.update(self._tvindex)
        self._ltid = tid

    def _abort(self):
        if self._nextpos:
            self._file.truncate(self._pos)
            self._nextpos=0

    def getSerial(self, oid):
        self._lock_acquire()
        try:
            return self._getSerial(oid, self._index[oid])
        finally:
            self._lock_release()

    def _getSerial(self, oid, pos):
        self._file.seek(pos)
        h = self._file.read(16)
        assert oid == h[:8]
        return h[8:]

    def _getVersion(self, oid, pos):
        # Return version and non-version pointer from oid record at pos.
        h = self._read_data_header(pos)
        if h.oid != oid:
            raise ValueError("invalid previous pointer")
        if h.version:
            return h.version, p64(h.pnv)
        else:
            return "", None

    def _undo_get_data(self, oid, pos, tpos):
        """Return the serial, data pointer, data, and version for the oid
        record at pos"""
        if tpos:
            pos = tpos - self._pos - self._thl
            tpos = self._tfile.tell()
            h = self._tfmt._read_data_header(pos)
            afile = self._tfile
        else:
            h = self._read_data_header(pos)
            afile = self._file
        if h.oid != oid:
            raise UndoError(oid, 'Invalid undo transaction id')

        refsdata = afile.read(h.nrefs * 8)
        if h.plen:
            data = afile.read(h.plen)
        else:
            data = ''
            pos = h.back

        if tpos:
            # Restore temp file to end
            self._tfile.seek(tpos)

        return h.serial, pos, data, h.version

    def _undo_record(self, h, pos):
        """Get the undo information for a data record

        Return a 6-tuple consisting of a pickle, references, data
        pointer, version, packed non-version data pointer, and current
        position.  If the pickle is true, then the data pointer must
        be 0, but the pickle can be empty *and* the pointer 0.
        """

        copy = 1 # Can we just copy a data pointer

        # First check if it is possible to undo this record.
        tpos = self._tindex.get(h.oid, 0)
        ipos = self._index.get(h.oid, 0)
        tipos = tpos or ipos

        if tipos != pos:
            # Eek, a later transaction modified the data, but,
            # maybe it is pointing at the same data we are.
            cserial, cdataptr, cdata, cver = self._undo_get_data(
                h.oid, ipos, tpos)
            # Versions of undone record and current record *must* match!
            if cver != h.version:
                raise UndoError(oid, 'Current and undone versions differ')

            if cdataptr != pos:
                # If the backpointers don't match, check to see if
                # conflict resolution is possible.  If not, raise
                # UndoError.
                try:
                    if (
                        # The current record wrote a new pickle
                        cdataptr == tipos
                        or
                        # Backpointers are different
                        self._loadBackPOS(h.oid, pos) !=
                        self._loadBackPOS(h.oid, cdataptr)
                        ):
                        if h.prev and not tpos:
                            copy = 0 # we'll try to do conflict resolution
                        else:
                            # We bail if:
                            # - We don't have a previous record, which should
                            #   be impossible.
                            raise UndoError(h.oid, "No previous record")
                except KeyError:
                    # LoadBack gave us a key error. Bail.
                    raise UndoError(h.oid, "_loadBack() failed")

        # Return the data that should be written in the undo record.
        if not h.prev:
            # There is no previous revision, because the object creation
            # is being undone.
            return "", "", 0, "", "", ipos

        # What does getVersion do?
        version, snv = self._getVersion(h.oid, h.prev)
        if copy:
            # we can just copy our previous-record pointer forward
            return "", "", h.prev, version, snv, ipos

        try:
            # returns data, serial tuple
            bdata = self._loadBack(h.oid, h.prev)[0]
        except KeyError:
            # couldn't find oid; what's the real explanation for this?
            raise UndoError(h.oid, "_loadBack() failed")


        # XXX conflict resolution needs to give us new references, but
        # that code isn't written yet

        try:
            data, refs = self._conflict.resolve(h.oid, cserial, h.serial,
                                                bdata, cdata)
        except interfaces.ConflictError:
            data = None
            refs = []

        if data is None:
            raise UndoError(h.oid,
                            "Some data were modified by a later transaction")
        return data, ''.join(refs), 0, h.version, snv, ipos


    # undoLog() returns a description dict that includes an id entry.
    # The id is opaque to the client, but contains the transaction id.
    # The transactionalUndo() implementation does a simple linear
    # search through the file (from the end) to find the transaction.

    def undoLog(self, first=0, last=-20, filter=None):
        if last < 0:
            last = first - last + 1
        self._lock_acquire()
        try:
            if self._packt is None:
                raise UndoError("Can't undo during pack")
            us = UndoSearch(self._file, self._pos, self._packt,
                            first, last, filter)
            while not us.finished():
                # Hold lock for batches of 20 searches, so default search
                # parameters will finish without letting another thread run.
                for i in range(20):
                    if us.finished():
                        break
                    us.search()
                # Give another thread a chance, so that a long undoLog()
                # operation doesn't block all other activity.
                self._lock_release()
                self._lock_acquire()
            return us.results
        finally:
            self._lock_release()

    def undo(self, transaction_id, transaction):
        """Undo a transaction, given by transaction_id.

        Do so by writing new data that reverses the action taken by
        the transaction.

        Usually, we can get by with just copying a data pointer, by
        writing a file position rather than a pickle. Sometimes, we
        may do conflict resolution, in which case we actually copy
        new data that results from resolution.
        """

        if self._is_read_only:
            raise ReadOnlyError()
        if transaction is not self._transaction:
            raise StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            return self._txn_undo(transaction_id)
        finally:
            self._lock_release()

    def _txn_undo(self, transaction_id):
        # Find the right transaction to undo and call _txn_undo_write().
        tid = base64.decodestring(transaction_id + '\n')
        assert len(tid) == 8
        tpos = self._txn_find(tid, 1)
        tindex = self._txn_undo_write(tpos)
        self._tindex.update(tindex)
        return tindex.keys()

    def _txn_find(self, tid, stop_at_pack):
        pos = self._pos
        # XXX Why 39?  Only because undoLog() uses it as a boundary.
        while pos > 39:
            self._file.seek(pos - 8)
            pos = pos - u64(self._file.read(8)) - 8
            self._file.seek(pos)
            h = self._file.read(TRANS_HDR_LEN)
            _tid = h[:8]
            if _tid == tid:
                return pos
            if stop_at_pack:
                # check the status field of the transaction header
                if h[16] == 'p' or _tid < self._packt:
                    break
        raise UndoError(None, "Invalid transaction id")

    def _txn_undo_write(self, tpos):
        # a helper function to write the data records for transactional undo

        otloc = self._pos
        here = self._pos + self._tfile.tell() + self._thl
        # Let's move the file pointer back to the start of the txn record.
        self._file.seek(tpos)
        h = self._file.read(TRANS_HDR_LEN)
        if h[16] != ' ':
            raise UndoError(None, "non-undoable transaction")
        tl = u64(h[8:16])
        ul, dl, el = struct.unpack(">HHH", h[17:TRANS_HDR_LEN])
        tend = tpos + tl
        pos = tpos + (TRANS_HDR_LEN + ul + dl + el)
        tindex = {}

        # keep track of failures, cause we may succeed later
        failures = {}
        while pos < tend:
            h = self._read_data_header(pos)
            if h.oid in failures:
                del failures[h.oid] # second chance!
            try:
                p, refs, prev, version, snv, ipos = self._undo_record(h, pos)
            except UndoError, v:
                # Don't fail right away. We may be redeemed later!
                failures[h.oid] = v
            else:
                plen = len(p)
                nrefs = len(refs) / 8
                self._tfile.write(pack(DATA_HDR,
                                       h.oid, self._serial, ipos, otloc,
                                       len(version), nrefs, plen))
                # If the backpointer refers to an object in a version,
                # we need to write a reasonable pointer to the previous
                # version data, which might be in _tvindex.
                if version:
                    vprev = (self._tvindex.get(version)
                             or self._vindex.get(version))
                    self._tfile.write(snv + p64(vprev) + version)
                    self._tvindex[version] = here
                    odlen = DATA_VERSION_HDR_LEN + len(version) + (plen or 8)
                else:
                    odlen = DATA_HDR_LEN + (plen or 8)

                self._tfile.write(refs)
                if p:
                    self._tfile.write(p)
                else:
                    self._tfile.write(p64(prev))
                tindex[h.oid] = here
                here += odlen + nrefs * 8
            pos += h.recordlen()
            if pos > tend:
                raise UndoError(None, "non-undoable transaction")
        if failures:
            raise MultipleUndoErrors(failures.items())

        return tindex

    def versionEmpty(self, version):
        if not version:
            # The interface is silent on this case. I think that this should
            # be an error, but Barry thinks this should return 1 if we have
            # any non-version data. This would be excruciatingly painful to
            # test, so I must be right. ;)
            raise interfaces.VersionError(
                "The version must be an non-empty string")
        self._lock_acquire()
        try:
            pos = self._vindex.get(version, 0)
            if not pos:
                return True
            while pos:
                h = self._read_data_header(pos)
                if self._index[h.oid] == pos:
                    return False
                pos = h.vprev
            return True
        finally:
            self._lock_release()

    def versions(self):
        return [version for version in self._vindex.keys()
                if not self.versionEmpty(version)]

    def pack(self, t, gc=True):
        """Perform a pack on the storage.

        There are two forms of packing: incremental and full gc.  In an
        incremental pack, only old object revisions are removed.  In a full gc
        pack, cyclic garbage detection and removal is also performed.

        t is the pack time.  All non-current object revisions older than t
        will be removed in an incremental pack.

        FileStorage ignores the gc flag.  Every pack does both incremental and
        full gc packing.
        """
        if self._is_read_only:
            raise ReadOnlyError()

        stop = timeStampFromTime(t).raw()
        if stop == ZERO:
            raise FileStorageError, 'Invalid pack time'

        # Record pack time so we don't undo while packing
        self._lock_acquire()
        try:
            # XXX _packt should just be a flag
            if self._packt != ZERO:
                # Already packing.
                raise FileStorageError, 'Already packing'
            # when undo sees _packt == None, it raises an error that
            # says "can't undo during pack"
            self._packt = None
        finally:
            self._lock_release()

        p = FileStoragePacker(self._name, stop,
                              self._lock_acquire, self._lock_release,
                              self._commit_lock_acquire,
                              self._commit_lock_release)
        try:
            t = p.pack()
            if t is None:
                return
            opos, index, vindex, tindex, tvindex = t
            oldpath = self._name + ".old"
            self._file.close()
            try:
                if os.path.exists(oldpath):
                    os.remove(oldpath)
                os.rename(self._name, oldpath)
            except Exception, msg:
                self._file = open(self._name, 'r+b')
                raise

            # OK, we're beyond the point of no return
            os.rename(self._name + '.pack', self._name)
            self._file = open(self._name, 'r+b')
            self._initIndex(index, vindex, tindex, tvindex)
            self._pos = opos
            self._save_index()
        finally:
            if p.locked:
                self._commit_lock_release()
                self._lock_release()
            self._lock_acquire()
            self._packt = ZERO
            self._lock_release()


    def iterator(self, start=None, stop=None):
        return FileIterator(self._file_name, start, stop)

    def lastTransaction(self):
        """Return transaction id for last committed transaction"""
        return self._ltid

    def lastSerial(self, oid):
        """Return last serialno committed for object oid.

        If there is no serialno for this oid -- which can only occur
        if it is a new object -- return None.
        """
        try:
            pos = self._index[oid]
        except KeyError:
            return None
        self._file.seek(pos)
        # first 8 bytes are oid, second 8 bytes are serialno
        h = self._file.read(16)
        if len(h) < 16:
            raise CorruptedDataError(oid, h)
        if h[:8] != oid:
            # get rest of header
            h += self._file.read(26)
            raise CorruptedDataError(oid, h)
        return h[8:]

    def cleanup(self):
        """Remove all files created by this storage."""
        cleanup(self._file_name)


class FileStoragePacker(FileStorageFormatter):

    def __init__(self, path, stop, la, lr, cla, clr):
        self._name = path
        self._file = open(path, "rb")
        self._stop = stop
        self.locked = False

        # The packer needs to acquire the parent's commit lock
        # during the copying stage, so the two sets of lock acquire
        # and release methods are passed to the constructor.
        self._lock_acquire = la
        self._lock_release = lr
        self._commit_lock_acquire = cla
        self._commit_lock_release = clr

    def _loada(self, oid, _index):
        """Read any version and return the version"""
        try:
            pos = _index[oid]
        except KeyError:
            raise POSKeyError(oid)
        h = self._read_data_header(pos)
        refs = splitrefs(self._file.read(h.nrefs * 8))
        if h.plen:
            data = self._file.read(h.plen)
        else:
            # _loadBack() leaves the file position sitting at the right place
            data = self._loadBack(oid, h.back)[0]
        return data, refs, h.version

    def _redundant_pack(self, pos):
        assert pos > 8, pos
        self._file.seek(pos - 8)
        p = u64(self._file.read(8))
        self._file.seek(pos - p + 8)
        return self._file.read(1) != ' ' # XXX or == "p"?

    def _pack_index(self, index):
        # Return an index for everything reachable at the pack time.
        rootl = [ZERO]
        pindex = fsIndex()
        while rootl:
            oid = rootl.pop()
            if oid in pindex:
                continue
            # XXX We might need to catch exceptions raised by _loada().
            p, refs, v = self._loada(oid, index)
            rootl.extend(refs)
            pindex[oid] = index[oid]
        return pindex

    def pack(self):
        packing = True

        ##################################################################
        # Step 1, get index as of pack time that has only referenced objects.
        index = fsIndex()
        vindex = {}
        # tindex is the index for the current transaction
        tindex = {}
        tvindex = {}
        packpos, maxoid, ltid = self._read_index(index, vindex, tindex,
                                                 self._stop, read_only=1)
        if packpos == self._metadata_size:
            return
        if self._redundant_pack(packpos):
            return
        pindex = self._pack_index(index)

        ##################################################################
        # Step 2, copy data and compute new index based on new positions.
        index = fsIndex()
        vindex = {}
        tindex = {}
        tvindex = {}

        # Index for non-version data.  This is a temporary structure
        # to reduce I/O during packing
        nvindex = fsIndex()

        ofile = open(self._name + '.pack', 'w+b')
        pv = ZERO
        offset = 0L  # the amount of space freed by packing
        pos = opos = self._metadata_size
        # Copy the metadata from the old file to the new one.
        self._file.seek(0)
        ofile.write(self._file.read(self._metadata_size))

        # Copy the data in two stages.  In the packing stage, we skip
        # records that are non-current or that are for unreferenced
        # objects. We also skip undone transactions.
        #
        # After the packing stage, we copy everything but undone
        # transactions, however, we have to update various back
        # pointers.  We have to have the storage lock in the second
        # phase to keep data from being changed while we're copying.
        pnv = None
        while True:
            # Check for end of packed records
            if packing and pos >= packpos:
                # OK, we're done with the old stuff, now we have to
                # get the lock so we can copy the new stuff!
                offset = pos - opos
                if offset <= 0:
                    # didn't free any space, there's no point in continuing
                    ofile.close()
                    self._file.close()
                    os.remove(self._name + '.pack')
                    return

                packing = False
                self._commit_lock_acquire()
                self._lock_acquire()
                self.locked = True

            # Read the transaction record
            self._file.seek(pos)
            h = self._file.read(TRANS_HDR_LEN)
            if len(h) < TRANS_HDR_LEN:
                break
            tid, tl, status, ul, dl, el = unpack(TRANS_HDR, h)
            if status == 'c':
                # Oops. we found a checkpoint flag.
                break
            tpos = pos
            tend = tpos + tl

            otpos = opos # start pos of output trans

            # write out the transaction record
            status = packing and 'p' or ' '
            ofile.write(h[:16] + status + h[17:])
            thl = ul + dl + el
            h = self._file.read(thl)
            if len(h) != thl:
                raise PackError(opos)
            ofile.write(h)
            thl = TRANS_HDR_LEN + thl
            pos = tpos + thl
            opos = otpos + thl

            while pos < tend:
                # Read the data records for this transaction
                h = self._read_data_header(pos)
                dlen = h.recordlen()
                if h.version:
                    if packing and pindex.get(h.oid, 0) != pos:
                        # This is not the most current record, or the
                        # oid is no longer referenced so skip it.
                        pos += dlen
                        continue

                    pv = p64(vindex.get(h.version, 0))
                    vindex[h.version] = opos
                else:
                    if packing:
                        # ppos and ph are the record pos and header as
                        # of the pack time.
                        ppos = pindex.get(h.oid, 0)
                        if ppos != pos:
                            if not ppos:
                                # object is no longer referenced, skip it.
                                pos += dlen
                                continue

                            # This is not the most current record, but
                            # maybe it's the most current committed record.
                            ph = self._read_data_header(ppos)
                            if not ph.version:
                                # The most current record is committed, so
                                # we can toss this one
                                pos += dlen
                                continue
                            self._file.read(ph.nrefs * 8)
                            pnv = self._loadBackPOS(h.oid, ph.pnv)
                            if pnv > pos:
                                # The current non version data is later,
                                # so this isn't the current record
                                pos += dlen
                                continue

                            # Ok, we've gotten this far, so we have the
                            # current record and we're ready to read the
                            # pickle, but we're in the wrong place, after
                            # wandering around to figure out if we were
                            # current.  Seek back to pickle data.
                            self._file.seek(pos + DATA_HDR_LEN)

                        nvindex[h.oid] = opos

                tindex[h.oid] = opos
                opos += dlen
                pos += dlen

                # Read the references
                refsdata = self._file.read(8 * h.nrefs)

                # Now load the pickle from the data record.  If there's a
                # backpointer, we might want to set plen, so copy into a local
                # to guarantee it's defined.
                plen = h.plen
                if plen:
                    p = self._file.read(h.plen)
                else:
                    if packing:
                        # When packing we resolve back pointers!
                        p, serial = self._loadBack(h.oid, h.back)
                        plen = len(p)
                        opos += plen - 8
                        splen = p64(plen)
                    else:
                        p = h.back
                        if p < packpos:
                            # We have a backpointer to a non-packed record.
                            # We have to be careful.  If we were pointing to a
                            # current record, then we should still point at
                            # one, otherwise, we should point at the last
                            # non-version record.
                            ppos = pindex.get(h.oid, 0)
                            if ppos:
                                if ppos == p:
                                    # we were pointing to the current record
                                    p = index[h.oid]
                                else:
                                    p = nvindex[h.oid]
                            else:
                                # Oops, this object was modified in a version
                                # in which it was deleted.  It doesn't matter
                                # what we use cause it's not reachable any
                                # more.
                                p = 0
                        else:
                            # This points back to a non-packed record.
                            # Just adjust for the offset
                            p -= offset
                        p = p64(p)

                prev = index.get(h.oid, 0)
                ofile.write(pack(DATA_HDR, h.oid, h.serial, prev, otpos,
                                 h.vlen, h.nrefs, plen))
                if h.version:
                    if not pnv:
                        ofile.write(ZERO)
                    else:
                        if pnv < packpos:
                            # we need to point to the packed non-version rec
                            pnv = nvindex.get(h.oid, 0)
                        else:
                            # we just need to adjust the pointer with
                            # the offset
                            pnv -= offset
                        ofile.write(p64(pnv))
                    ofile.write(pv)
                    ofile.write(h.version)
                ofile.write(refsdata)
                ofile.write(p)

            # skip the (intentionally redundant) transaction length
            pos += 8

            if self.locked:
                # temporarily release the lock to give other threads
                # a chance to do some work!
                self._commit_lock_release()
                self._lock_release()
                self.locked = False

            index.update(tindex) # Record the position
            tindex.clear()

            # Now, maybe we need to hack or delete the transaction
            otl = opos - otpos
            if otl != tl:
                # Oops, what came out is not what came in!

                # Check for empty:
                if otl == thl:
                    # Empty, slide back over the header:
                    opos = otpos
                    ofile.seek(opos)
                else:
                    # Not empty, but we need to adjust transaction length
                    # and update the status
                    ofile.seek(otpos + 8)
                    otl = p64(otl)
                    ofile.write(otl + status)
                    ofile.seek(opos)
                    ofile.write(otl)
                    opos += 8

            else:
                ofile.write(p64(otl))
                opos += 8

            if not packing:
                # We are in the copying phase. We need to get the lock
                # again to avoid someone writing data while we read it.
                self._commit_lock_acquire()
                self._lock_acquire()
                self.locked = True

        # OK, we've copied everything. Now we need to wrap things up.
        ofile.flush()
        ofile.close()
        self._file.close()

        return opos, index, vindex, tindex, tvindex

def _truncate(file, name, pos):
    seek=file.seek
    seek(0,2)
    file_size=file.tell()
    try:
        i=0
        while 1:
            oname='%s.tr%s' % (name, i)
            if os.path.exists(oname):
                i=i+1
            else:
                warn("Writing truncated data from %s to %s", name, oname)
                o=open(oname,'wb')
                seek(pos)
                cp(file, o, file_size-pos)
                o.close()
                break
    except:
        error("couldn\'t write truncated data for %s", name)
        raise StorageSystemError("Couldn't save truncated data")

    seek(pos)
    file.truncate()

class FileIterator(FileStorageFormatter):
    """Iterate over the transactions in a FileStorage file."""
    _ltid = ZERO
    _file = None

    __implements__ = IStorageIterator

    def __init__(self, filename, start=None, stop=None):
        self._file = open(filename, "rb")
        self._read_metadata()
        self._file.seek(0,2)
        self._file_size = self._file.tell()
        self._pos = self._metadata_size
        assert start is None or isinstance(start, str)
        assert stop is None or isinstance(stop, str)
        if start:
            self._skip_to_start(start)
        self._stop = stop

    def close(self):
        file = self._file
        if file is not None:
            self._file = None
            file.close()

    def _skip_to_start(self, start):
        # Scan through the transaction records doing almost no sanity
        # checks.
        while True:
            self._file.seek(self._pos)
            h = self._file.read(16)
            if len(h) < 16:
                return
            tid, stl = unpack(">8s8s", h)
            if tid >= start:
                return
            tl = u64(stl)
            try:
                self._pos += tl + 8
            except OverflowError:
                self._pos = long(self._pos) + tl + 8
            if __debug__:
                # Sanity check
                self._file.seek(self._pos - 8, 0)
                rtl = self._file.read(8)
                if rtl != stl:
                    pos = self._file.tell() - 8
                    panic("%s has inconsistent transaction length at %s "
                          "(%s != %s)",
                          self._file.name, pos, u64(rtl), u64(stl))

    def __iter__(self):
        if self._file is None:
            # A closed iterator.  XXX: Is IOError the best we can do?  For
            # now, mimic a read on a closed file.
            raise IOError("iterator is closed")
        file = self._file
        seek = file.seek
        read = file.read

        pos = self._pos
        while True:
            # Read the transaction record
            seek(pos)
            h = read(TRANS_HDR_LEN)
            if len(h) < TRANS_HDR_LEN:
                break

            tid, tl, status, ul, dl, el = unpack(TRANS_HDR,h)
            if el < 0:
                el = t32 - el

            if tid <= self._ltid:
                warn("%s time-stamp reduction at %s", self._file.name, pos)
            self._ltid = tid

            if pos+(tl+8) > self._file_size or status=='c':
                # Hm, the data were truncated or the checkpoint flag wasn't
                # cleared.  They may also be corrupted,
                # in which case, we don't want to totally lose the data.
                warn("%s truncated, possibly due to damaged records at %s",
                     self._file.name, pos)
                break

            if status not in ' up':
                warn('%s has invalid status, %s, at %s', self._file.name,
                     status, pos)

            if tl < (TRANS_HDR_LEN+ul+dl+el):
                # We're in trouble. Find out if this is bad data in
                # the middle of the file, or just a turd that Win 9x
                # dropped at the end when the system crashed.  Skip to
                # the end and read what should be the transaction
                # length of the last transaction.
                seek(-8, 2)
                rtl = u64(read(8))
                # Now check to see if the redundant transaction length is
                # reasonable:
                if self._file_size - rtl < pos or rtl < TRANS_HDR_LEN:
                    nearPanic('%s has invalid transaction header at %s',
                              self._file.name, pos)
                    warn("It appears that there is invalid data at the end of "
                         "the file, possibly due to a system crash.  %s "
                         "truncated to recover from bad data at end.",
                         self._file.name)
                    break
                else:
                    warn('%s has invalid transaction header at %s',
                         self._file.name, pos)
                    break

            if self._stop is not None and tid > self._stop:
                return

            tpos = pos
            tend = tpos+tl

            if status == 'u':
                # Undone transaction, skip it
                seek(tend)
                l = u64(read(8))
                if l != stl:
                    panic('%s has inconsistent transaction length at %s',
                          self._file.name, pos)
                pos = tend+8
                continue

            pos = tpos+(TRANS_HDR_LEN+ul+dl+el)
            user = read(ul)
            description = read(dl)
            e = {}
            if el:
                try:
                    e = loads(read(el))
                # XXX can we do better?
                except:
                    pass

            result = RecordIterator(tid, status, user, description, e, pos,
                                    tend, file, tpos)
            pos = tend

            # Read the (intentionally redundant) transaction length
            seek(pos)
            l = u64(read(8))
            if l != tl:
                warn("%s redundant transaction length check failed at %s",
                     self._file.name, pos)
                break
            pos += 8
            yield result

class RecordIterator(FileStorageFormatter):
    """Iterate over the transactions in a FileStorage file."""

    __implements__ = ITransactionRecordIterator, ITransactionAttrs

    def __init__(self, tid, status, user, desc, ext, pos, tend, file, tpos):
        self.tid = tid
        self.status = status
        self.user = user
        self.description = desc
        self._extension = ext
        self._pos = pos
        self._tend = tend
        self._file = file
        self._tpos = tpos

    def __iter__(self):
        pos = self._pos
        while pos < self._tend:
            # Read the data records for this transaction
            h = self._read_data_header(pos)
            dlen = h.recordlen()
            if pos + dlen > self._tend or h.tloc != self._tpos:
                warn("%s data record exceeds transaction record at %s",
                     file.name, pos)
                return

            pos += dlen
            prev_txn = None

            if h.plen:
                refsdata = self._file.read(h.nrefs * 8)
                refs = splitrefs(refsdata)
                data = self._file.read(h.plen)
            else:
                if not h.back:
                    # If the backpointer is 0, then this transaction
                    # undoes the object creation.  It either aborts
                    # the version that created the object or undid the
                    # transaction that created it.  Return None
                    # for data and refs because the backpointer has
                    # the real data and refs.
                    data = None
                    refs = None
                else:
                    data, refs, _s, tid = self._loadBackTxn(h.oid, h.back)
                    prev_txn = self.getTxnFromData(h.oid, h.back)

            yield Record(h.oid, h.serial, h.version, data, prev_txn, refs)

class Record:
    """An abstract database record."""

    __implements__ = IDataRecord

    def __init__(self, oid, serial, version, data, data_txn, refs):
        self.oid = oid
        self.serial = serial
        self.version = version
        self.data = data
        self.data_txn = data_txn
        self.refs = refs

class UndoSearch:

    def __init__(self, file, pos, packt, first, last, filter=None):
        self.file = file
        self.pos = pos
        self.packt = packt
        self.first = first
        self.last = last
        self.filter = filter
        self.i = 0
        self.results = []
        self.stop = 0

    def finished(self):
        """Return True if UndoSearch has found enough records."""
        # BAW: Why 39 please?  This makes no sense (see also below).
        return self.i >= self.last or self.pos < 39 or self.stop

    def search(self):
        """Search for another record."""
        dict = self._readnext()
        if dict is not None and (self.filter is None or self.filter(dict)):
            if self.i >= self.first:
                self.results.append(dict)
            self.i += 1

    def _readnext(self):
        """Read the next record from the storage."""
        self.file.seek(self.pos - 8)
        self.pos -= u64(self.file.read(8)) + 8
        self.file.seek(self.pos)
        h = self.file.read(TRANS_HDR_LEN)
        tid, tl, status, ul, dl, el = struct.unpack(TRANS_HDR, h)
        if tid < self.packt or status == 'p':
            self.stop = 1
            return None
        if status != ' ':
            return None
        d = u = ''
        if ul:
            u = self.file.read(ul)
        if dl:
            d = self.file.read(dl)
        e = {}
        if el:
            try:
                e = loads(self.file.read(el))
            except:
                pass
        d = {'id': base64.encodestring(tid).rstrip(),
             'time': TimeStamp(tid).timeTime(),
             'user_name': u,
             'description': d}
        d.update(e)
        return d

class DataHeader:
    """Header for a data record."""

    __slots__ = (
        "oid", "serial", "prev", "tloc", "vlen", "plen", "nrefs", "back",
        # These three attributes are only defined when vlen > 0
        "pnv", "vprev", "version")

    version = ""
    back = 0

    def __init__(self, oid, serial, prev, tloc, vlen, nrefs, plen):
        self.oid = oid
        self.serial = serial
        self.prev = prev
        self.tloc = tloc

        self.vlen = vlen
        self.nrefs = nrefs
        self.plen = plen

    def fromString(cls, s):
        return cls(*struct.unpack(DATA_HDR, s))

    fromString = classmethod(fromString)

    def parseVersion(self, buf):
        self.pnv, self.vprev = struct.unpack(">QQ", buf[:16])
        self.version = buf[16:]

    def recordlen(self):
        rlen = DATA_HDR_LEN + (self.nrefs * 8) + (self.plen or 8)
        if self.version:
            rlen += 16 + self.vlen
        return rlen


def cleanup(filename):
    """Remove all FileStorage related files."""
    for ext in '', '.old', '.tmp', '.lock', '.index', '.pack':
        try:
            os.remove(filename + ext)
        except OSError, e:
            if e.errno != errno.ENOENT:
                raise
