##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import os
import shutil
import tempfile
import unittest
from StringIO import StringIO

import ZConfig.Context

from zodb import config


class StorageTestBase(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.tmpfn = tempfile.mktemp()
        self.storage = None

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        storage = self.storage
        self.storage = None
        try:
            if storage is not None:
                storage.close()
        except:
            pass
        try:
            # Full storage creates a directory
            if os.path.isdir(self.tmpfn):
                shutil.rmtree(self.tmpfn)
            else:
                os.remove(self.tmpfn)
        except os.error:
            pass

    def loadConfigText(self, text):
        context = ZConfig.Context.Context()
        io = StringIO(text)
        return context.loadFile(io)


class StorageTestCase(StorageTestBase):
    def testFileStorage(self):
        from zodb.storage.file import FileStorage
        sample = """
        <Storage>
        type       FileStorage
        file_name  %s
        create     yes
        </Storage>
        """ % self.tmpfn
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, FileStorage)
        self.assertEqual(args, {"file_name": self.tmpfn, "create": 1})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, FileStorage))

    def testZEOStorage(self):
        try:
            from zodb.zeo.client import ClientStorage
        except ImportError:
            return
        sample = """
        <Storage>
        type       ClientStorage
        addr       zeo://www.python.org:9001
        wait       no
        </Storage>
        """
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, ClientStorage)
        self.assertEqual(args, {"addr": [("www.python.org", 9001)], "wait": 0})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, ClientStorage))

    def testMappingStorage(self):
        from zodb.storage.mapping import MappingStorage
        sample = """
        <Storage>
        type       MappingStorage
        </Storage>
        """
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, MappingStorage)
        self.assertEqual(args, {})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, MappingStorage))

    def testModuleStorage(self):
        # Test explicit module+class
        from zodb.storage.mapping import MappingStorage
        sample = """
        <Storage>
        type       zodb.storage.mapping.MappingStorage
        </Storage>
        """
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, MappingStorage)
        self.assertEqual(args, {})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, MappingStorage))


class BerkeleyTestCase(StorageTestBase):
    def testFullStorage(self):
        from zodb.storage.bdbfull import BDBFullStorage
        sample = """
        <Storage>
        type       BDBFullStorage
        name       %s
        cachesize  1000
        </Storage>
        """ % self.tmpfn
        os.mkdir(self.tmpfn)
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, BDBFullStorage)
        # It's too hard to test the config instance equality
        args = args.copy()
        del args['config']
        self.assertEqual(args, {"name": self.tmpfn})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, BDBFullStorage))
        # XXX _config isn't public
        self.assert_(self.storage._config.cachesize, 1000)

    def testMinimalStorage(self):
        from zodb.storage.bdbminimal import BDBMinimalStorage
        sample = """
        <Storage>
        type       BDBMinimalStorage
        name       %s
        cachesize  1000
        </Storage>
        """ % self.tmpfn
        os.mkdir(self.tmpfn)
        rootconf = self.loadConfigText(sample)
        storageconf = rootconf.getSection("Storage")
        cls, args = config.getStorageInfo(storageconf)
        self.assertEqual(cls, BDBMinimalStorage)
        # It's too hard to test the config instance equality
        args = args.copy()
        del args['config']
        self.assertEqual(args, {"name": self.tmpfn})
        self.storage = config.createStorage(storageconf)
        self.assert_(isinstance(self.storage, BDBMinimalStorage))
        # XXX _config isn't public
        self.assert_(self.storage._config.cachesize, 1000)


test_classes = [StorageTestCase]

from zodb.storage.base import berkeley_is_available
if berkeley_is_available:
    test_classes.append(BerkeleyTestCase)


def test_suite():
    suite = unittest.TestSuite()
    for klass in test_classes:
        sub = unittest.makeSuite(klass)
        suite.addTest(sub)
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
